package projects.dream5;

import java.io.File;
import java.io.PrintWriter;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.classifier.scoringFunctionBased.gendismix.GenDisMixClassifier;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.Sample;
import de.jstacs.data.Sequence;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.ParameterSetTagger;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.scoringFunctions.NormalizedScoringFunction;
import de.jstacs.scoringFunctions.directedGraphicalModels.MutableMarkovModelScoringFunction;
import de.jstacs.scoringFunctions.homogeneous.HMMScoringFunction;
import de.jstacs.scoringFunctions.mix.StrandScoringFunction;
import de.jstacs.scoringFunctions.mix.motifSearch.HiddenMotifsMixture;
import de.jstacs.utils.Pair;



public class DreamPredictPublic {
	
	private static String[] keys = {"home","file","model","outfile"};
	
	private static class DreamPredictPublicParameterSet extends ParameterSet{

		@Override
		protected void loadParameters() throws Exception {
			initParameterList();
			parameters.add( new SimpleParameter(DataType.STRING, "home directory", "The home directory where the data reside.",true,".") );
			parameters.add( new SimpleParameter(DataType.STRING, "input file", "The input file in Dream5 format for predictions (column 1: TF, column 2: array type, column 3: sequence), one file per TF and array type, path relative to home directory.", true ) );
			parameters.add( new SimpleParameter( DataType.STRING, "model", "File with the trained model stored as XML, path relative to home directory.",true,"model.xml") );
			parameters.add( new SimpleParameter( DataType.STRING, "outfile", "File where the predictions are stored.",true,"predictions.txt") );
		}
		
	}
	

	public static void main( String[] args ) throws Exception {
		DreamPredictPublicParameterSet ps = new DreamPredictPublicParameterSet();
		ParameterSetTagger pst = new ParameterSetTagger( keys, ps );
		pst.fillParameters( "=", args );
		if(!pst.hasDefaultOrIsSet()){
			System.err.println("Some parameter values missing");
			System.err.println(pst.toString());
			System.exit( 1 );
		}
		System.out.println(pst.toString());
		
		String home = (String) pst.getValueFromTag( keys[0] ), inputFile = (String) pst.getValueFromTag( keys[1] ), classifier = (String) pst.getValueFromTag( keys[2] ), outfile = (String) pst.getValueFromTag( keys[3] );
		
		Pair<String[], Sample> pair = getData( home, inputFile );
		
		String tf = pair.getFirstElement()[0];
		String array = pair.getFirstElement()[1];
		
		Sample data = pair.getSecondElement();
	
		GenDisMixClassifier cl = new GenDisMixClassifier( FileManager.readFile( new File(home+System.getProperty( "file.separator" )+classifier) ) );

		HiddenMotifsMixture zoops = (HiddenMotifsMixture) cl.getScoringFunction(0);
		int fgOrder = ((MutableMarkovModelScoringFunction)((StrandScoringFunction)((NormalizedScoringFunction)zoops.getFunction(0)).getFunction()).getFunction(0)).getOrder();
		int flOrder=((HMMScoringFunction)zoops.getFunction(2)).getMaximalMarkovOrder();
		System.out.println( "Model: "+classifier + "\nMotif order: "+fgOrder+"\nFlanking order: "+flOrder+"\nScore: "+cl.getLastScore());
		System.out.println("Writing predicitions to "+outfile);

		PrintWriter wr = new PrintWriter( home+System.getProperty( "file.separator" )+outfile );
		
		double[] pred = Evaluation.getPred(cl,data.getInfixSample(0, 40));	
		for(int i=0;i<data.getNumberOfElements();i++){
			wr.println(tf+"\t"+array+"\t"+data.getElementAt( i )+"\t"+pred[i]);
		}
		
		wr.close();
		System.out.println("finished.");			
					
	}
	
	
	public static Pair<String[], Sample> getData(String home, String file ) throws Exception {
		
		AlphabetContainer con = new AlphabetContainer( new DNAAlphabet() );
		SparseStringExtractor exAn = new SparseStringExtractor( home+System.getProperty( "file.separator" )+file,'#' );
		LinkedList<Sequence> list = new LinkedList<Sequence>();
		
		String lastTF = "";
		
		while(exAn.hasMoreElements()){
			String line = exAn.nextElement();
			String[] temp = line.split( "\t" );
			Sequence seq = Sequence.create( con, temp[2] );			
			
			if((!lastTF.equals( "" ) && !lastTF.equals( temp[0] )) || !exAn.hasMoreElements()){
				
				return new Pair<String[], Sample>(new String[]{temp[0], temp[1]},new Sample( "", list.toArray( new Sequence[0] ) ));
				
				
			}
			list.add( seq );
			lastTF = temp[0];
		}
		
		return null;
		
	}
	
}
