package projects.dream5;

import java.io.File;
import java.util.Arrays;
import java.util.LinkedList;

import de.jstacs.DataType;
import de.jstacs.algorithms.optimization.ConstantStartDistance;
import de.jstacs.algorithms.optimization.Optimizer;
import de.jstacs.classifier.scoringFunctionBased.OptimizableFunction.KindOfParameter;
import de.jstacs.classifier.scoringFunctionBased.gendismix.GenDisMixClassifier;
import de.jstacs.classifier.scoringFunctionBased.gendismix.GenDisMixClassifierParameterSet;
import de.jstacs.classifier.scoringFunctionBased.gendismix.LearningPrinciple;
import de.jstacs.classifier.scoringFunctionBased.gendismix.LogGenDisMixFunction;
import de.jstacs.classifier.scoringFunctionBased.gendismix.OneSampleLogGenDisMixFunction;
import de.jstacs.classifier.scoringFunctionBased.logPrior.CompositeLogPrior;
import de.jstacs.classifier.scoringFunctionBased.logPrior.LogPrior;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.Sample;
import de.jstacs.data.Sequence;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.io.ArrayHandler;
import de.jstacs.io.FileManager;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.motifDiscovery.MutableMotifDiscovererToolbox;
import de.jstacs.motifDiscovery.MutableMotifDiscovererToolbox.InitMethodForScoringFunction;
import de.jstacs.motifDiscovery.history.CappedHistory;
import de.jstacs.motifDiscovery.history.History;
import de.jstacs.motifDiscovery.history.NoRevertHistory;
import de.jstacs.parameters.ParameterSet;
import de.jstacs.parameters.ParameterSetTagger;
import de.jstacs.parameters.SimpleParameter;
import de.jstacs.scoringFunctions.NormalizableScoringFunction;
import de.jstacs.scoringFunctions.NormalizedScoringFunction;
import de.jstacs.scoringFunctions.directedGraphicalModels.MutableMarkovModelScoringFunction;
import de.jstacs.scoringFunctions.directedGraphicalModels.structureLearning.measures.InhomogeneousMarkov;
import de.jstacs.scoringFunctions.homogeneous.HMMScoringFunction;
import de.jstacs.scoringFunctions.mix.StrandScoringFunction;
import de.jstacs.scoringFunctions.mix.StrandScoringFunction.InitMethod;
import de.jstacs.scoringFunctions.mix.motifSearch.DurationScoringFunction;
import de.jstacs.scoringFunctions.mix.motifSearch.HiddenMotifsMixture;
import de.jstacs.scoringFunctions.mix.motifSearch.MixtureDuration;
import de.jstacs.scoringFunctions.mix.motifSearch.SkewNormalLikeScoringFunction;
import de.jstacs.scoringFunctions.mix.motifSearch.UniformDurationScoringFunction;
import de.jstacs.utils.ComparableElement;
import de.jstacs.utils.DoubleList;
import de.jstacs.utils.Pair;
import de.jstacs.utils.SafeOutputStream;



public class DreamPublic {
	
	private static String[] keys = {"home","file","mo","fo","starts","threads","q","model"};
	
	private static class DreamPublicParameterSet extends ParameterSet{

		@Override
		protected void loadParameters() throws Exception {
			initParameterList();
			parameters.add( new SimpleParameter(DataType.STRING, "home directory", "The home directory where the data reside.",true,".") );
			parameters.add( new SimpleParameter(DataType.STRING, "input file", "The input file in Dream5 format (column 1: TF, column 2: array type, column 3: sequences, column 4: signal, last column: flag), one file per TF and array type, path relative to home directory.", true ) );
			parameters.add( new SimpleParameter(DataType.INT, "motif order", "The order of the inhomogeneous Markov model for the motif.",true,1) );
			parameters.add( new SimpleParameter( DataType.INT, "flanking oder", "The order of the homogeneous Markov model for flanking sequence and background.", true,3 ) );
			parameters.add( new SimpleParameter( DataType.INT, "starts", "The number of starts of the optimization.", true,5 ) );
			parameters.add( new SimpleParameter( DataType.INT, "threads", "The number of threads, i.e. cores, that are used for optimization.", true,1 ) );
			parameters.add( new SimpleParameter( DataType.DOUBLE, "q", "A-priori fraction of data points with weight greater than 0.5", true, 0.025 ) );
			parameters.add( new SimpleParameter( DataType.STRING, "model", "File where the trained model is stored as XML, path relative to home directory.",true,"model.xml") );
		}
		
	}

	public static void main( String[] args ) throws Exception {
		
		DreamPublicParameterSet ps = new DreamPublicParameterSet();
		ParameterSetTagger pst = new ParameterSetTagger( keys, ps );
		pst.fillParameters( "=", args );
		if(!pst.hasDefaultOrIsSet()){
			System.err.println("Some parameter values missing");
			System.err.println(pst.toString());
			System.exit( 1 );
		}
		System.out.println(pst.toString());
		
			
			String home = (String) pst.getValueFromTag( keys[0] );
			String inputFile = (String) pst.getValueFromTag( keys[1] );
			int fgOrder = (Integer) pst.getValueFromTag( keys[2] );
			int bgOrder = (Integer) pst.getValueFromTag( keys[3] );
			int restarts = (Integer) pst.getValueFromTag( keys[4] );
			int threads = (Integer) pst.getValueFromTag( keys[5] );
			boolean bgsEq = true;
			double p = (Double) pst.getValueFromTag( keys[6] );
			String modelXML = (String) pst.getValueFromTag( keys[7] );
			
			Interpolation inter = Interpolation.RANK_LOG;
			int motifL = 8;//TODO
			double eps = 1E-7, lineps = 1E-10, startD = 1;//TODO
			//double eps = 1, lineps = 1, startD = 1;//TODO
			byte algo = Optimizer.QUASI_NEWTON_BFGS;
			SafeOutputStream stream = new SafeOutputStream( System.out );
			History history = new CappedHistory( 15, new NoRevertHistory(true,false,false) );
			InitMethodForScoringFunction[] initMeth = {InitMethodForScoringFunction.PLUG_IN, InitMethodForScoringFunction.NOTHING};
			int anz = 250;
			
			Pair<double[], Sample> pair = getWeightedData( home, inputFile );
			Sample data = pair.getSecondElement();
			data = data.getInfixSample(0, 40);
			System.out.println("Number of sequences: "+data.getNumberOfElements());
			
			double[] signal = pair.getFirstElement();
			double[] fgWeights = Interpolation.getWeight( data, signal, p, inter );
			double[][] weights = { fgWeights, Interpolation.getBgWeight(fgWeights) };
						
			
			int sl = data.getElementLength();
			double ess = 5;//TODO 20;
			LogPrior prior = new CompositeLogPrior();
			double[] beta = LearningPrinciple.getBeta( LearningPrinciple.MSP );
			
			int sampleSize = 100;
			
			NormalizableScoringFunction[] score = new NormalizableScoringFunction[2];
			
			double[] bgHyperPars = new double[bgsEq ? bgOrder+1 : 1];
			for(int i=0;i<bgHyperPars.length-1;i++){
				bgHyperPars[i] = ess*(1-p)/p;
			}
			bgHyperPars[bgHyperPars.length-1] = ess*(sl - (bgsEq ? bgOrder : 0) )*(1-p)/p;
			score[1] = new HMMScoringFunction(data.getAlphabetContainer(),bgsEq ? bgOrder : 0,ess*(1-p)/p,bgHyperPars,true,true,1);
			
			LogGenDisMixFunction objective, initObjective;
					
			// create scoring function with hidden motif
			SkewNormalLikeScoringFunction motifPenalty = new SkewNormalLikeScoringFunction( 1, 50, -1.5, -2.5, 4.2 );
			//System.out.println( motifPenalty );
			System.out.println();
			HMMScoringFunction flanking = null;
			double[] flHyperPars = new double[bgOrder+1];
			for(int i=0;i<bgOrder;i++){
				flHyperPars[i] = ess*0.2;
			}
			flHyperPars[flHyperPars.length-1] = ess*(sl-bgOrder-0.2*motifL);
			
			flanking = new HMMScoringFunction(data.getAlphabetContainer(),bgOrder,ess*0.2,flHyperPars,true,true,1);
			
			double essMotif = ess*0.8;
			NormalizableScoringFunction motif = 
				new MutableMarkovModelScoringFunction( data.getAlphabetContainer(), motifL, essMotif, true, new InhomogeneousMarkov(fgOrder), motifPenalty );
			motif = new StrandScoringFunction( motif, 0.5, 1, true, InitMethod.INIT_FORWARD_STRAND );
			motif.initializeFunctionRandomly( false );
			motif = new NormalizedScoringFunction( motif, 1 );
			
			DurationScoringFunction pos = null;
			double seqs = 1, sd = 20, skewSd = 1, mu = 18;
			pos = new MixtureDuration( 1,
						new UniformDurationScoringFunction( 0, sl-motifL, essMotif-seqs ),
						new SkewNormalLikeScoringFunction( 0, sl-motifL, true, mu, 500, true, seqs/2d, seqs/2d*(sd*sd), true, 0, skewSd, 1) );
			score[0] = 
				new HiddenMotifsMixture( HiddenMotifsMixture.CONTAINS_SOMETIMES_A_MOTIF, sl, 1, true, flanking, motif, pos, true );
	
			LearningPrinciple initKey = (beta[LearningPrinciple.CONDITIONAL_LIKELIHOOD_INDEX]>0) ? LearningPrinciple.MCL : LearningPrinciple.ML;	
			
			double[][] res, best = null;
					
			double[][] smallWeights = new double[weights.length][];
			Sample[] dd = {null,null};
			LinkedList<Sequence> list = new LinkedList<Sequence>();
			DoubleList wList = new DoubleList();
			double[] help = weights[0].clone();
			
			
			System.out.println("for init:");
			for( int i = 0; i < weights.length; i++ ) {
				list.clear();
				wList.clear();
				System.arraycopy(weights[i], 0, help, 0, help.length);
				Arrays.sort(help);
				double t = help[help.length-sampleSize];
				//System.out.println( help[0] + "\t" + t + "\t" + help[help.length-1] );
				for( int j = 0; j < weights[i].length; j++ ) {
					if( weights[i][j] >= t && list.size() < sampleSize ) {
						list.add( data.getElementAt(j) );
						wList.add( weights[i][j] );
					}
				}
				smallWeights[i] = wList.toArray();
				dd[i] = new Sample( "", list.toArray( new Sequence[0]) );
				
				//System.out.println( "class " + i + "\t" + dd[i].getNumberOfElements() + " sequences" );
			}
			initObjective = new LogGenDisMixFunction( threads, score, dd, smallWeights, prior, LearningPrinciple.getBeta( initKey ), true, false );
			objective = new OneSampleLogGenDisMixFunction( threads, score, data, weights, prior, beta, true, false );
			
			GenDisMixClassifierParameterSet cps = new GenDisMixClassifierParameterSet( data.getAlphabetContainer(), sl, algo, eps, lineps, startD, false, KindOfParameter.PLUGIN, true, threads );
			NormalizableScoringFunction[] current, bestNSF = null;
			for( int r = 0; r < restarts; r++ ) {
				System.out.println( "start " + r + "\n-------------------------------------------------" );
				current = ArrayHandler.clone( score );		
				
				objective.reset( current );
				initObjective.reset( current );
				
				// initialize
				ComparableElement<double[], Double>[] pars = MutableMotifDiscovererToolbox.getSortedInitialParameters( current, initMeth, initObjective, anz, stream, 0 );
				objective.setParams( pars[pars.length-1].getElement() );
				
				//System.out.println( current[0] );
				System.out.println("_________________________________");
		
				// optimize
				objective.reset( current );
				res = MutableMotifDiscovererToolbox.optimize( current, objective, algo, eps, lineps, new ConstantStartDistance(startD),
						stream, false, history, KindOfParameter.PLUGIN, true );
				
				// select
				if( best == null || res[0][0] > best[0][0] ) {
					bestNSF = current;
					best = res;
				}
			}
			// save classifier
			
			GenDisMixClassifier cl = new GenDisMixClassifier( cps, prior, best[0][0], beta, bestNSF );
			cl.setClassWeights( false, best[1] );
			//TODO save
			FileManager.writeFile( new File( home+System.getProperty( "file.separator" )+modelXML), cl.toXML() );
			System.out.println( cl );
			
			System.out.println( "optimized: " + best[0][0] );
				
			
		
	}

	
	public static Pair<double[],Sample> getWeightedData(String home, String file ) throws Exception {
		
		AlphabetContainer con = new AlphabetContainer( new DNAAlphabet() );
		SparseStringExtractor exAn = new SparseStringExtractor( home+System.getProperty( "file.separator" )+file,'#' );
		LinkedList<Sequence> list = new LinkedList<Sequence>();
		DoubleList truth = new DoubleList();
		String lastTF = "";
		
		while(exAn.hasMoreElements()){
			String line = exAn.nextElement();
			String[] temp = line.split( "\t" );
			Sequence seq = Sequence.create( con, temp[2] );
			double weight = Double.NaN;
			if(!"NA".equals( temp[3] ) ){
				weight = Double.parseDouble( temp[3] );
			}
			int flag = Integer.parseInt( temp[temp.length-1] );
			
			
			if((!lastTF.equals( "" ) && !lastTF.equals( temp[0] )) || !exAn.hasMoreElements()){
				
				Sample data = new Sample( "", list.toArray( new Sequence[0] ) );
				double[] signal = truth.toArray();
				
				return new Pair<double[], Sample>( signal, data );
				
			}
			if(flag == 0){
				list.add( seq );
				truth.add( weight );
			}
			lastTF = temp[0];
		}
		
		return null;
		
	}
	
}
