package projects.dream5;

import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;

import de.jstacs.classifier.AbstractScoreBasedClassifier;
import de.jstacs.classifier.ScoreBasedPerformanceMeasureDefinitions;
import de.jstacs.data.Sample;
import de.jstacs.data.Sequence;
import de.jstacs.io.SparseStringExtractor;
import de.jstacs.utils.DoubleList;
import de.jstacs.utils.IntList;
import de.jstacs.utils.ToolBox;

/**
 * 
 * @author Jan Grau
 */
public class Evaluation {

	public static enum Mapping{
		DIRECT,
		NEAREST_NEIGHBOUR,
		MEAN,
		MEDIAN,
		WEIGHTED
	}
	
	public static int getNumberOfResults() {
		return 6;
	}
	
	public static String getHeading() {
		return "spearman\tpearson\tpearson log\tauc-PR\tauc-ROC\tCC_8";
	}
	
	public static void fillResults(double[] pred, double[] signal, Sample data, LinkedList<Sequence> positives, DoubleList results) throws Exception{
		results.add( ToolBox.spearmanCorrelation(pred, signal) );
		results.add( ToolBox.pearsonCorrelation(pred,signal) );
		double[] logPred = new double[pred.length], logSignal = logPred.clone();
		for( int i = 0; i < pred.length; i++ ) {
			logPred[i] = Math.log( pred[i] );
			logSignal[i] = Math.log( signal[i] );
		}
		results.add( ToolBox.pearsonCorrelation(logPred,logSignal) );
		KMerStatistics kmerStat = new KMerStatistics(data,signal,true,0,8,false);
		kmerStat.fillStatistics( pred, positives, results );
	}
	
	public static void printResults( DoubleList results ) {
		System.out.print( results.get(0) );
		for( int i = 1; i < results.length(); i++ ) {
			System.out.print( "\t" + results.get(i) );	
		}
		System.out.println();
	}
	
	public static double[] getPred( AbstractScoreBasedClassifier cl, Sample data ) throws Exception{
		double[] pred = cl.getScores( data );
		
		for(int i=0;i<pred.length;i++){
			pred[i] = 1.0/(1.0 + Math.exp( -pred[i] ));//TODO koennten wir uns evtl auch sparen, da monoton
		}
		return pred;
	}
	
	public static double[] transformScores2(String home, double[] pred, double[] train, double[][][][] intensities, int array, int tf, Mapping mapping) throws Exception {
		train = train.clone();
		Arrays.sort( train );
		double[] result = null;
		SparseStringExtractor ex  = null;
		
		int maxLength = 0;
		for(int k=0;k<intensities[1][1-array].length;k++){
			if( k!=tf && intensities[1][1-array][k].length>maxLength){
				maxLength = intensities[1][1-array][k].length;
			}
		}
		
		double[] transformed = new double[maxLength];
		int[] ns = new int[maxLength];
		switch(mapping){
			case MEAN:
				for(int k=0;k<intensities[1][1-array].length;k++){
					if(k!=tf) {
						for( int i = intensities[1][1-array][k].length-1, j = maxLength-1; i >= 0; i--, j-- ) {
							transformed[j] += intensities[1][1-array][k][i];
							ns[j]++;
						}
					}
				}
				
				for(int i=0;i<transformed.length;i++){
					transformed[i] /= ns[i];
				}
				break;
			case MEDIAN:
				double[] values = null;
				for(int k=0;k<intensities[1][1-array].length;k++){
					if(k!=tf) {
						for( int i = intensities[1][1-array][k].length-1, j = maxLength-1; i >= 0; i--, j-- ) {
							ns[j]++;
						}
					}
				}
				for(int z = 0, i=transformed.length-1;i>=0;i--, z++){
					if( values == null || values.length != ns[i] ) {
						values = new double[ns[i]];
					}
					
					int a = ns[i];
					for(int k=0;k<intensities[1][1-array].length;k++){
						if(k!=tf) {
							int idx= intensities[1][1-array][k].length-1-z;
							if( idx >= 0 ) {
								a--;
								//System.out.println( i + "\t" + ns[i] + "\t" + a + "\t" + k + "\t" + idx );
								values[a] = intensities[1][1-array][k][idx];
							}
						}
					}					
					
					Arrays.sort( values );
					if(values.length%2==0){
						transformed[i] = (values[(int)Math.floor( (ns[i]-1)/2.0 )] + values[(int)Math.ceil( (ns[i]-1)/2.0 )])/2.0;
					}else{
						transformed[i] = values[(ns[i]-1)/2];
					}
				}
				break;
			case NEAREST_NEIGHBOUR:
			case WEIGHTED:
				int idx = -1;
				double best = Double.POSITIVE_INFINITY, curr = 0;
				DoubleList rmses = new DoubleList();
				IntList idxs = new IntList();
				for(int k=0;k<intensities[1][array].length;k++){
					if(k!=tf) {
						curr = rmse(train,intensities[1][array][k]);
						rmses.add( curr );
						idxs.add( k );
						if( curr > 0 && curr < best){//TODO
							best = curr;
							idx = k;
						}
					}
				}
				//System.out.println("best: "+best+", "+idx);

				if(mapping == Mapping.NEAREST_NEIGHBOUR){
					transformed=intensities[1][1-array][idx];
				}else{
					double[] weights = rmses.toArray();
					double min = ToolBox.min( weights );
					double max = ToolBox.max( weights );
					for(int i=0;i<weights.length;i++){
						weights[i] = max - (weights[i] - min);
					}
				//  System.out.println(Arrays.toString( weights ));
					for(int g = 0, k=0;k<intensities[1][1-array].length;k++, g++){
						if(k!=tf){
							for(int i=intensities[1][1-array][k].length-1, j=maxLength-1;i>=0;i--,j--){
								transformed[j] += weights[g]*intensities[1][1-array][k][i];
								ns[j]+=weights[g];
							}
						} else {
							g--;
						}
					}

					for(int i=0;i<transformed.length;i++){
						transformed[i] /= ns[i];
					}
				}
				break;
			case DIRECT:
				transformed = train;
				break;
			default:
				throw new Exception();
		}
		result = transformScores( pred, transformed );
		return result;
	}
	
	public static double[] transformScores(double[] pred, double[] trainIntens){
		int[] predRanks = ToolBox.rank( pred, false );
		trainIntens = trainIntens.clone();
		Arrays.sort( trainIntens );
		double[] transformed = new double[pred.length];
		double fac = (double)trainIntens.length/(double) pred.length;
		for(int i=0;i<transformed.length;i++){
			int upIdx = (int)Math.ceil( fac*predRanks[i] );
			if(upIdx >= trainIntens.length){upIdx = trainIntens.length-1;}
			int downIdx = (int)Math.floor( fac*predRanks[i] );
			if(downIdx < 0){downIdx = 0;}
			double p = fac*predRanks[i] - downIdx;
			
			transformed[i] = p*trainIntens[trainIntens.length-1-upIdx] + (1.0-p)*trainIntens[trainIntens.length-1-downIdx];
		}
		return transformed;
	}
	
	public static double rmse(double[] pred, double[] truth){
		double rmse = 0;
		double n = Math.min( pred.length, truth.length );
		for(int i=0;i<n;i++){
			rmse += (pred[(int)n-i-1] - truth[(int)n-i-1])*(pred[(int)n-i-1] - truth[(int)n-i-1])/n;
		}
		return Math.sqrt( rmse );
	}
}
