package projects.dream5;

import java.util.Arrays;
import java.util.BitSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;

import de.jstacs.data.Sample;
import de.jstacs.data.Sequence;
import de.jstacs.io.ArrayHandler;
import de.jstacs.motifDiscovery.KMereStatistic;
import de.jstacs.utils.DoubleList;
import de.jstacs.utils.ToolBox;

public enum Interpolation {
	LOG_LINEAR,
	LINEAR_jens,
	LINEAR,
	RANK_OUTLIER,
	RANK_OUTLIER_REPEATED,
	RANK_LOG,
	RANK_SIGMOID,
	SIGMOID;
	
	public static double[] getWeight( Sample data, double[] intensa, double aPrioriWeight, Interpolation interpol ) throws Exception {
		double[] weights = new double[intensa.length];
		
		double min, max, thresh;
		double[] help = intensa.clone();
		Arrays.sort( help );
		min = help[0];
		max = help[help.length-1];
		thresh = help[ (int) Math.ceil( (1.0-aPrioriWeight)*help.length )];
		
		double a1, b1, c = 1E-10, h, m=0, q25, q75, temp;
		int idx = -1;
		int[] rank;
		switch(interpol) {
			case LOG_LINEAR:
				min = Math.log(min);
				max = Math.log(max)-min;
				for(int i=0;i<intensa.length;i++){
					weights[i] = aPrioriWeight+(1-2*aPrioriWeight)*(Math.log(intensa[i]) - min)/max;
				}
				break;
			case LINEAR_jens:
				max-=min;
				for(int i=0;i<intensa.length;i++){
					weights[i] = (intensa[i] - min)/max;
				}
				break;
			case LINEAR:
				double distAbove = max - thresh;
				double distBelow = thresh - min;
				for(int i=0;i<intensa.length;i++){
					if(intensa[i] >= thresh){
						temp = intensa[i] - thresh;
						weights[i] = 0.5 + 0.5*temp/distAbove;
					}else{
						temp = thresh - intensa[i];
						weights[i] = 0.5 - 0.5*temp/distBelow;
					}
				}
				break;
			case RANK_OUTLIER_REPEATED:
			case RANK_OUTLIER:
				boolean repeat = interpol == RANK_OUTLIER_REPEATED;
				int n1, n2, idx1 = 0, idx2 = -1000;
				do {
					if( idx2 < 0 ) {
						n1 = 0;
						n2 = help.length;
					} else {
						n1 = idx1;
						n2 = idx2;
					}
					
					q25 = help[n1+(int)(0.25*(n2-n1))];
					q75 = help[n1+(int)(0.75*(n2-n1))];
					/*
					thresh = q25 - 3*(q75-q25);
					idx1 = Arrays.binarySearch( help, thresh );
					if( idx1 < 0 ) {
						idx1 = -idx1 + 1;
					}
					*/
					thresh = q75 + 1.5*(q75-q25);
					idx2 = Arrays.binarySearch( help, thresh );
					if( idx2 < 0 ) {
						idx2 = -idx2 + 1;
					}
					//System.out.println( help[n1] + "\t" + thresh + "\t" + help[n2-1] + "\t" + idx1 + "\t" + n2 );
				} while( repeat && idx2 < n2 );
				aPrioriWeight = 1 - idx2 / (double) intensa.length;
				System.out.println( aPrioriWeight );
			case RANK_LOG:
				rank = ToolBox.rank(intensa,true);
				for(int i=0;i<intensa.length;i++){
					if( m <= rank[i] ) {
						m = rank[i];
					}
				}
				for(int i=0;i<intensa.length;i++){
					/*
					h = 1d-(rank[i] / m);
					weights[i] = 1d/(1d + (1-h)*(1-aPrioriWeight) / (h*aPrioriWeight));//TODO changed aPrioriWeight
					weights[i] = 1d/(1d + (1-h)/h * (1-aPrioriWeight)/aPrioriWeight );
					*/
					
					h = rank[i] / m;
					weights[i] = 1d/(1d + h/(1-h) * (1-aPrioriWeight)/aPrioriWeight);
				}
				break;
			case SIGMOID:
				b1 = (Math.log( c ) - Math.log1p( -c ))/(1.0 - max/thresh);
				a1 = b1/thresh;
				/*
				double b2 = (Math.log1p( -c ) - Math.log( c ))/(1.0 - min/thresh);
				double a2 = b2/thresh;
				if(a2 < a1){
					a1 = a2;
					b1 = b2;
				}*/
				for(int i=0;i<intensa.length;i++){
					weights[i] = 1.0/(1.0 + Math.exp( -a1*intensa[i] + b1 ));
				}
				break;
			case RANK_SIGMOID://TODO Jan
				rank = ToolBox.rank(intensa,true);
				int[] rankSorted = rank.clone();
				Arrays.sort( rankSorted );
				
				min = -rankSorted[rankSorted.length-1];
				max = -rankSorted[0];
				thresh = -rankSorted[ (int) Math.floor( (aPrioriWeight)*rankSorted.length )];
				System.out.println(min+" "+max+" "+thresh);
				
				b1 =  (Math.log1p( -c ) - Math.log( c ))/(1.0 - min/thresh);
				a1 = b1/thresh;
				/*
				double b2 = (Math.log1p( -c ) - Math.log( c ))/(1.0);
				double a2 = b2/thresh;
				if(a2 < a1){
					a1 = a2;
					b1 = b2;
				}*/
				for(int i=0;i<intensa.length;i++){
					weights[i] = 1.0/(1.0 + Math.exp( -a1*(-rank[i]) + b1 ));
				}
				break;
			default:
				/*
				AlphabetContainer contcont = new AlphabetContainer( new ContinuousAlphabet() );
				GaussianEmission em1 = new GaussianEmission( contcont );
				GaussianEmission em2 = new GaussianEmission( contcont );

				Sequence[] intenss = new Sequence[intensa.length];
				for(int i=0;i<intensa.length;i++){
					intenss[i] = new ArbitrarySequence( contcont, new double[]{intensa[i]} );
					if(intensa[i] >= thresh){
						em1.addToStatistic( true, 0, 0, 1, intenss[i] );
					}else{
						em2.addToStatistic( true, 0, 0, 1, intenss[i] );
					}
				}
				
				em1.estimateFromStatistic();
				em2.estimateFromStatistic();
				System.out.println(em1);
				System.out.println(em2);
				
				
				double logAPriori = Math.log( aPrioriWeight );
				double log1mAPriori = Math.log1p( -aPrioriWeight );
				
				double[] temp = new double[2];
				for(int i=0;i<intensa.length;i++){
					temp[0] = logAPriori + em1.getLogProbFor( true, 0, 0, intenss[i] );
					temp[1] = log1mAPriori + em2.getLogProbFor( true, 0, 0, intenss[i] );
					Normalisation.logSumNormalisation( temp );
					weights[1][0][i] = temp[0];
					weights[1][1][i] = temp[1];
				}
				*/
				throw new RuntimeException();
		}
		
		return weights;
	}
		
	public static double[] getBgWeight( double[] weights ) {
		double[] bgWeights = new double[weights.length];
		for(int i=0;i<bgWeights.length;i++){
			bgWeights[i] = 1.0 - weights[i];			
		}
		return bgWeights;
	}
}
